/*
 * OpenEAS Experlog Application Server
 * Web application development and dynamic template scripting engine
 * Copyright (c) 2004 ExperLog SARL - http://www.experlog.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version. This program is distributed in
 * the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
 * the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have
 * received a copy of the GNU General Public License along with this program;
 * if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package org.experlog.base;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Properties;

import javax.servlet.ServletConfig;
import javax.servlet.http.HttpServletRequest;

import org.apache.log4j.Category;
import org.experlog.db.ESConnectionPool;
import org.experlog.util.Client;
import org.experlog.util.ConnectionPool;
import org.experlog.util.ESCategory;
import org.experlog.util.AppConfiguration;
import org.experlog.util.Store;
import org.experlog.util.ESTimer;
import org.experlog.util.ESTimerThread;

/**
 * @author spengler
 *
 * An online store.
 * this object contains all the client from a store
 */
public class ESStore implements Store, ESTimer {

    private Hashtable clientPool_ = new Hashtable();	// pool of (clientID, ESClient)
    public ConnectionPool connectionPool_;

    private AppConfiguration shopConfig_ = null;
    private String shopName_ = "openeas";
    private int state_ = 1;
    private int max = 0;
    private int maxClients_ = 0;
    private int countClients_ = 0;
    private long timeoutmillis_ = 3600000;
    private ESTimerThread timer_;
    private Category logger;

    private String confdir_ = null;
    private String conffile_;
    private Properties general_ = new Properties();
    
    private static ESStore theStore_ = null;

    public static synchronized ESStore getInstance() {
      return getInstance(null);
    }
    public static synchronized ESStore getInstance(ServletConfig cfg) {
      if(theStore_ == null) theStore_ = new ESStore(cfg);
      return theStore_;
    }
    
    public ESStore(ServletConfig cfg) {
      if(logger != null) logger.debug("shopName=<" + shopName_ + ">");
      if(logger != null) logger.debug("conf=<" + conffile_ + "> \nconfdir_=<" + confdir_
          + "> \nconffile_=<" + conffile_ + ">");
      try {

        configure(cfg);

        shopConfig_ = new AppConfiguration(shopName_, general_, true, conffile_, confdir_);
        
        try {
          // try to load the connection pool object (put in the configuration
          // file)
          connectionPool_ = new ESConnectionPool();
        } catch (Exception e) {
          if(logger != null) logger.error("ESStore impossible to create a new pool of connections",
              e);
        }
        if (shopConfig_ == null)
          if(logger != null) logger.debug("shopConfig_ is null name=<openeas>");
          else {
            if(logger != null) logger.debug("shopConfig_ is not null name=<openeas>");
            timeoutmillis_ = shopConfig_.getTimeout() * 1000;
          }
        connectionPool_.setShopConfig(shopConfig_, "openeas");

        timer_ = new ESTimerThread(this, timeoutmillis_ / (2*1000), null);
        timer_.start();
        maxClients_ = shopConfig_.getMaxClients();
        
      } catch (Exception e) {
        System.err.println(e.toString());
        e.printStackTrace(System.err);
      }
    }
    
    private synchronized void configure(ServletConfig cfg) {
      logger = ESCategory.getCategory("openeas.store");
      if(logger != null) logger.debug("configure");

      confdir_ = cfg.getInitParameter("SHOPDIR");
      if (confdir_ == null)
        confdir_ = cfg.getInitParameter("shopdir");

      String conf = cfg.getInitParameter("CONFIGFILE");
      if (conf == null)
        conf = cfg.getInitParameter("configfile");
      if (conf == null) {
        conf =
          (onWindows()
              ? new String("examples/lite/DemoWin.cfg")
          : new String("examples/lite/Demo.cfg"));
      }

      FileInputStream f = null;
      try {
        StringBuffer conffile = new StringBuffer();
        if (confdir_ != null) {
          conffile.append(confdir_);
          if (!confdir_.endsWith(File.separator))
            conffile.append(File.separator);
        }
        conffile.append(conf);

        conffile_ = conffile.toString();

        if (!new File(conffile_).isAbsolute()) {
          confdir_ = cfg.getServletContext().getRealPath("/");
          if (confdir_ != null) {
            if (!confdir_.endsWith(File.separator))
              confdir_ += File.separator;
            conffile_ = confdir_ + conffile_;
          }
        }

        f = new FileInputStream(conffile_);
        general_.load(f);
        f.close();
        f = null;

      } catch (IOException e) {
        if (f != null)
          try {
            f.close();
          } catch (Exception ee) {
            if(logger != null) logger.error("configure fail to close file :" + ee);
          }
          if(logger != null) logger.error("configure IO Exception :" + e);
      }

    }

    protected static boolean onWindows() {
      return File.separatorChar == '\\';
    }

    public static synchronized String makeId(String name, boolean unique) {
      long val;

      if (name == null) {
        name = new String("");
        unique = true;
      }

      if (unique) {
        val = System.currentTimeMillis();
        while (val == System.currentTimeMillis());
      } else {
        val = (long) Math.ceil(Math.random() * 100000 - 1);
      }

      String s = (name.length() > 4 ? name.substring(0, 4) : name);
      String base = s.trim().replace('\t', '-').replace(' ', '-');

      return new String(base + val);
    }

    private String getClientId(String id) {

      if (id == null) {
        return new String(
            getShopConfiguration().getName() + "__" + makeId(null, true));
      } else if (id.indexOf("__") < 0) {
        return new String(id + "__" + makeId(null, true));
      } else
        return id;
    }

    /**
     * get an client id from the servlet request if the client already exists,
     * returns it or creates a new one, creates a new jdbc connection and store
     * it in the clientPool object
     * 
     * if the client exists, just set the cookie
     * 
     * Called by ESServlet (upon HTTP request)
     */
    public OpenClient getClient(String id, HttpServletRequest req)
    throws Exception {
      String clientId = getClientId(id);
      OpenClient ret = (OpenClient) clientPool_.get(clientId);
      if(ret != null) {
        ret.getCookie().setCookie(req);
        return ret;
      }
      else return syncGetClient(clientId, req);
    }

    private synchronized OpenClient syncGetClient(String clientId, HttpServletRequest req)
    throws Exception {

      OpenClient ret;
      if (((ret = (OpenClient) clientPool_.get(clientId)) == null)
          && (state_ == 1)) {
        if (countClients_ < maxClients_) {
          ret = new OpenClient(clientId, req, connectionPool_);
          countClients_ ++;
          clientPool_.put(clientId, ret);
          if (clientPool_.size() > max)
            max = clientPool_.size();
        } else return null;
      } else if ((ret = (OpenClient) clientPool_.get(clientId)) != null) {
        // Client already in pool
        ret.getCookie().setCookie(req);
      } else { // state != 0
        return null;
      }
      return ret;
    }

    /**
     * get an ESClient object from the clientPool pool
     */
    public OpenClient findClient(String clientId) {
    	return (OpenClient) clientPool_.get(clientId);
    }

    /**
     * get all the ESClient objects from the clientPool pool
     */
    public Enumeration listClients() {
      return clientPool_.keys();
    }

    /**
     * Drop a specified client: releases all resources allocated for that
     * client (database connection,  etc...)
     * @param cid the ID of the client to drop
     */
    public void endClient(String cid) {
      if(logger != null) logger.debug("ESStore:endClient");
      OpenClient c = (OpenClient) clientPool_.remove(cid);
      c.endClient();
      countClients_ --;
    }

    /**
     * returns the shop configuration for the current shop
     */
    public AppConfiguration getShopConfiguration() {
      return shopConfig_;
    }

    public int getClientPool() {
      if (clientPool_ != null)
        return clientPool_.size();
      else
        return 0;
    }

    public String getConPool() {
      return connectionPool_.toString();
    }

    public String toString() {
      return "ESStore:toString { shopName=["
      + shopName_
      + "], "
      + "clientPool=["
      + getClientPool()
      + "], "
      + "cnxPool=["
      + getConPool()
      + "] }";
    }

    public void setShopName(String shop) {
      shopName_ = shop;
    }

    public void setClientPoolTable(Hashtable table) {
      clientPool_ = table;
    }

    public void setConnectionPool(ConnectionPool conn) {
      connectionPool_ = conn;
    }

    public void setShopConfig(AppConfiguration conf) {
      shopConfig_ = conf;
    }

    public void setShopProps(Properties shopProps) {
      general_ = shopProps;
    }

    public String getShopName() {
      return shopName_;
    }
    public Hashtable getClientPoolTable() {
      return clientPool_;
    }

    public ConnectionPool getConnectionPool() {
      return connectionPool_;
    }

    public AppConfiguration getShopConfig() {
      return shopConfig_;
    }

    public Properties getShopProps() {
      return general_;
    }

    public int getState() {
      return state_;
    }

    public void setState(int state) {
	if (state == -1) {
	    for (Enumeration enu = clientPool_.keys();
		 enu.hasMoreElements();
		 ) {
		String cl = (String) enu.nextElement();
		endClient(cl);
	    }
	    //	    connectionPool_.cleanup();
	}

	state_ = state;
    }

    public int getMaxClients() {
      return max;
    }

    public void setLevel(String newLevel) {
      ESCategory.setLevel(newLevel, shopName_);
    }

    public String getLevel() {
      return ESCategory.getLevel(shopName_);
    }

    public String getOutput() {
      return ESCategory.getOutput(shopName_);
    }
    
    public String getLogRequest() {
        if(logger != null) logger.debug("ESStore: getLogRequest call");
        String result = new String();
        
        for (Enumeration enu = clientPool_.elements(); enu.hasMoreElements();) {
    		 OpenClient ec = (OpenClient) enu.nextElement();
                 Hashtable lr = ec.getLogRequest();
                 for (Enumeration enulog = lr.elements(); enulog.hasMoreElements();) {
                         DatasourceRequest dr = (DatasourceRequest) enulog.nextElement();
                         result += dr.toString();
                 }
    	    }
        if(logger != null) logger.debug("ESStore: getLogRequest ["+getShopName()+"] call result=("+result+")");
        return result;
    }
    
    public void tick(Object info) {
      Enumeration keys = clientPool_.keys();
      while (keys.hasMoreElements()) {
        String cid = (String) keys.nextElement();
        if (((Client)clientPool_.get(cid)).sessionExpired())
          endClient(cid);
      }
	  timer_ = new ESTimerThread(this, timeoutmillis_ / (2*1000), null);
	  try {
	    Runtime.getRuntime().gc();
	  } catch (Exception te) {
	    if(logger != null) logger.error("ESStore:tick() - gc problem :" + te);
	  }
	  timer_.start();
    }
}
