/*
 * OpenEAS Experlog Application Server
 * Web application development and dynamic template scripting engine
 * Copyright (c) 2004 ExperLog SARL - http://www.experlog.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version. This program is distributed in
 * the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
 * the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have
 * received a copy of the GNU General Public License along with this program;
 * if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package org.experlog.db;

import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Hashtable;
import java.util.Iterator;

import org.apache.log4j.Category;
import org.experlog.util.Configuration;
import org.experlog.util.ConnectionPool;
import org.experlog.util.ESCategory;
import org.experlog.util.AppConfiguration;

import org.experlog.openeas.api.ConnectionManager;
import org.enhydra.jdbc.pool.GenerationObject;

/**
 * ESConnectionPool is a database tool object.
 * it is used by openeas to do some specific operation on
 * jdbc object. The main property of this object is the ESDataSourceWrapper
 * object which is a wrapper to a DataSource.
 */
public class ESConnectionPool implements ConnectionPool {

    // Database connexions configuration

    private boolean new_ = true;
    // true = the jdbc driver has not yet been loaded
    private AppConfiguration shopConfig_ = null;
    private int qryattempts_ = 1;
    private boolean reconnect_on_err_ = false; // reconnect DB on request error
    private ESDataSourceWrapper dataSource_; // a wrapper to a dataSource
    private Category logger;
    private Hashtable datasources;

    /**
     * constructor of the connection pool object
     * NO WORK is done here
     */
    public ESConnectionPool() {
    }

    /**
     * set the configuration of the shop, and before the call of this
     * method.
     */
    public void setShopConfig(Configuration shopConfig, String name)
	throws Exception {
	logger = ESCategory.getCategory(name, "openeas.db");
	logger.debug(
		     "ESConnectionPool:setShopConfig set the shop config for "
		     + "this object");

	shopConfig_ = (AppConfiguration)shopConfig;

	// the shopconfig is now known, we can create the dataSource wrapper
	if (dataSource_ == null) {
	    dataSource_ = new ESDataSourceWrapper(name);

	    dataSource_.init(shopConfig_, null);
            ConnectionManager.addDataSource(dataSource_);
	    if (shopConfig != null) {
		datasources = new Hashtable();
		Hashtable htable = shopConfig.getDatasources();
		if (htable != null) {
		    logger.debug("There are datasources to configure, "
				 + "configure them now");

		    Iterator iter = htable.values().iterator();
		    while (iter.hasNext()) {
			DatasourceProperty dsprop = (DatasourceProperty) iter.next();

			logger.debug("add the following datasource=<"
				     + dsprop.getName() + ">");

			ESDataSourceWrapper esds = new ESDataSourceWrapper(name);

			logger.debug("init the ESDataSourceWrapper with the "
				     + "shopConfig and the DatasourceProperty");
			esds.init(shopConfig_, dsprop);
                        ConnectionManager.addDataSource(esds);

			esds.setMinCon(dsprop.getMinConn());
			esds.setMaxCon(dsprop.getMaxConn());

			datasources.put(dsprop.getName(), esds);

			logger.debug(
				     "add the new ESDataSourceWrapper to the "
				     + "Administration center");
		    }
		}
	    }

	} else {
	    logger.error(
			 "ESConnectionPool:setShopConfig ERROR: try to set "
			 + "again the shop config property");
	}

	dataSource_.setMinCon(shopConfig_.getMinConn());
	dataSource_.setMaxCon(shopConfig_.getMaxConn());
    }

    /**
     * load the JDBC driver by a class forName method.
     * and return a connection
     */
    public synchronized Connection connectDb(String datasourcename) throws Exception {
	logger.debug("ESConnectionPool:connectionDb load the Jdbc driver ");

	Connection c = null;
		
	// Incredible but true:
	// Some JDBC drivers require we open/close a connection once
	// before they work (!)
	if (new_) {
	    String open2 = shopConfig_.get("jdbc.PreOpen");
	    if (open2 != null && open2.equalsIgnoreCase("true")) {
		if (datasourcename == null) {
		    logger.debug("connectionDb just get a connection");

		    c =	dataSource_.getConnection(shopConfig_.getJdbcUser(),
						  shopConfig_.getJdbcPassword());

		    logger.debug("connectionDb just release this connection (close)");

		    c.close();
		} else {
		    logger.debug("New_, the datasource is <" + datasourcename + ">");
		    Hashtable htable = shopConfig_.getDatasources();
		    DatasourceProperty esds =
			(DatasourceProperty) htable.get(datasourcename);

		    c = ((ESDataSourceWrapper) datasources.get(datasourcename)).
			getConnection(esds.getUser(),
				      esds.getPassword());

		    logger.debug("New_ end, the connection is released");

		    c.close();
		}
	    }
	    new_ = false;
	}
	logger.debug(
		     "ESConnectionPool:connectionDb call the getConnection method on DataSource");

	// try to have a connection from the pool
	if (datasourcename == null){
	    logger.debug("the name of the datasource is null, default status");
	    c =	dataSource_.getConnection(shopConfig_.getJdbcUser(),
					  shopConfig_.getJdbcPassword());
	    setTxIsolation(c, shopConfig_.getJdbcTxIsolation());
	} else {
	    logger.debug("the name of the datasource is not null, try to get "
			 + "a new connection from it");

	    Hashtable htable = shopConfig_.getDatasources();
	    DatasourceProperty esds = (DatasourceProperty) htable.get(datasourcename);

	    logger.debug("get a connection from a datasource (hashtable) <"
			 + datasourcename
			 + "> from client=<"
			 + esds.getUser()
			 + ">");

	    c =	((ESDataSourceWrapper) datasources.get(datasourcename)).
		getConnection(esds.getUser(),
			      esds.getPassword());

	    setTxIsolation(c, esds.getTxIsolation());
	}
	if (c == null)
	    logger.debug(
			 "ESConnectionPool:connectionDb return an invalid connection");
	else
	    logger.debug(
			 "ESConnectionPool:connectionDb return a valid connection");

	return c;
    }

    /**
     * Update the store's database
     */
    public void updateDB(Connection con, String sql)
	throws SQLException {
	logger.debug("ESConnectionPool:updateDb update the store\'s database");
	Statement st = null;

	try {
	    st = con.createStatement();
	} catch (SQLException nc) {
	    logger.error("["
			 + new java.util.Date(System.currentTimeMillis())
			 + "] ERROR IN createStatement, sql=["
			 + sql
			 + "] ",
			 nc);
	    throw(nc);
	}

	logger.debug("ESConnectionPool:updateDB SESSION updateDB: sql=" + sql);

	try {
	    st.executeUpdate(sql);
	} catch (SQLException nc) {
	    logger.error(
			 "["
			 + new java.util.Date(System.currentTimeMillis())
			 + "] ERROR IN executeUpdate, sql=["
			 + sql
			 + "] ",
			 nc);
	    throw(nc);
	}
	finally {
	    if (st != null)
		st.close();
	}
    }

    public int getQryattempts() {
	return qryattempts_;
    } // method getQryattempts

    public boolean isReconnect_on_err() {
	return reconnect_on_err_;
    } // method isReconnect_on_err

    public boolean isNew() {
	return new_;
    } // method isNew

    public ESDataSourceWrapper getDataSource() {
	return dataSource_;
    } // method getDataSource

    public void setQryattempts(int qryattempts) {
	qryattempts_ = qryattempts;
    } // method setQryattempts

    public void setReconnect_on_err(boolean reconnect) {
	reconnect_on_err_ = reconnect;
    } // method setReconnect_on_err

    public void setNew(boolean newb) {
	new_ = newb;
    } // method setNew

    public synchronized void cleanup() {
	dataSource_.cleanup();
	if (datasources != null) {
	    Iterator iter = datasources.values().iterator();
	    while (iter.hasNext()) {
		ESDataSourceWrapper esdw = (ESDataSourceWrapper) iter.next();
		esdw.cleanup();
	    }
	}
    }

    private void setTxIsolation(Connection c, int txisol) {
	if (txisol == AppConfiguration.INT_DEFAULT || c == null)
	    return;
	try {
	    logger.debug("Setting connection\'s TX isolation to: " + txisol);
	    c.setTransactionIsolation(txisol);
	} catch (SQLException e) { // Ignore
	    logger.debug("Warning: Could not set TX isolation (ignored).");
	}
    }

    public Hashtable getPool() {
	return dataSource_.getPool();
    }

    public void removeUnlockedObject(GenerationObject obj) {
	dataSource_.removeUnlockedObject(obj);
    }

    public Hashtable getUsedPool() {
	return dataSource_.getUsedPool();
    }

    public void removeLockedObject(GenerationObject obj) {
	dataSource_.removeLockedObject(obj);
    }
}