/*
 * OpenEAS Experlog Application Server
 * Web application development and dynamic template scripting engine
 * Copyright (c) 2004 ExperLog SARL - http://www.experlog.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version. This program is distributed in
 * the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
 * the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have
 * received a copy of the GNU General Public License along with this program;
 * if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package org.experlog.gencode;

import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.sql.ResultSet;
import java.util.Collection;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.StringTokenizer;
import java.util.Vector;

import org.apache.log4j.Category;
import org.experlog.db.ESDBDataSet;
import org.experlog.openeas.api.VectorDataSet;

/**
 * UserObject is an object to manage Java objects calls. This object is used by
 * ESDynHtml to create object with :
 *   $NewObject myObject $TypeOfObject([param]*)$
 * and to call methods on object with :
 *   $Invoke resultObject $myObject.myMethod([param]*)$
 * Parameters are separated by comma.
 */
public class UserObject {

  /**
   * name_ is the name of the object ($NewObject name_ ...)
   */
  String name_;

  /**
   * aclass_ is the Class object of the object
   */
  Class aclass_;

  /**
   * obj_ is the object itself, and used to call methods on it
   */
  Object obj_;

  /**
   * type_ is the type of the object (only its string representation)
   */
  String type_;

  /**
   * just a logger to log output string
   */
  Category logger_ = null;

  String sessionId_;
  boolean persist_ = false;

  public UserObject(Object obj, String name, Category logger) throws Exception {
    obj_ = obj;
    name_ = name;
    logger_ = logger;
    try {
      if(obj == null) { // name will be considered a class name
        type_ = name;
        aclass_ = Class.forName(name);
      } else {
        type_ = obj_.getClass().getName();
        aclass_ = Class.forName(type_);
      }
    } catch (Exception e) {
      logger_.debug("UserObject(" + obj + "," + name + "): can\'t guess class name, default java.lang.Object: " + e);
      // we do not have the name of the object type, we use Object instead
      type_ = "java.lang.Object";
      aclass_ = Class.forName(type_);
    }
  }

  /**
   * create a new UserObject, user asks for a new Java object to create
   *
   * @param line  the line which comes from the template
   * @param lnb  the number of the line in the template
   * @param uObjects  an Hashtable objects which contains all UserObjects
         * @param sObjects      an Hashtable objects which contains sessions UserObjects
   * 
   * @throws Exception
   */
  public UserObject(String line, int lnb, Hashtable uObjects, Hashtable sObjects, Category logger)
    throws Exception {
                logger_ = logger;
                
    // remove blank
    line = line.trim();

    StringTokenizer st = new StringTokenizer(line);
    st.nextToken(); //Skip $NewObject

    if (!st.hasMoreTokens())
      throw new IOException(syntaxError(line, lnb));

                String persist = st.nextToken();
                if (persist.equalsIgnoreCase("isPersist")) {
                        // we must create a new Persistent UserObject
                        persist_ = true;
                        name_ = st.nextToken();
                } else
                        name_ = persist;
                        
    // defining an object without its constructor is not possible
    if (!st.hasMoreTokens())
      throw new IOException(syntaxError(line, lnb));

    String c =
      line
        .substring(line.indexOf(name_) + name_.length(), line.length())
        .trim();
    c = c.substring(1, c.length() - 1);
    String className = c.substring(0, c.indexOf('('));

    // need to be saved for future use
    type_ = className;

    // get the list of parameters with '(' and ')'
    String listParams = c.substring(c.indexOf('(') + 1, c.lastIndexOf(')'));
    logger_.debug(
      "className =<" + className + ">, listParams=<" + listParams + ">");

    if (className.compareTo("String") == 0)
      className = "java.lang.String";
    if (className.compareTo("Boolean") == 0)
      className = "java.lang.Boolean";
    if (className.compareTo("Byte") == 0)
      className = "java.lang.Byte";
    if (className.compareTo("Character") == 0)
      className = "java.lang.Character";                        
    if (className.compareTo("Class") == 0)
      className = "java.lang.Class";
    if (className.compareTo("Double") == 0)
      className = "java.lang.Double";
    if (className.compareTo("Float") == 0)
      className = "java.lang.Float";
    if (className.compareTo("Integer") == 0)
      className = "java.lang.Integer";
    if (className.compareTo("Long") == 0)
      className = "java.lang.Long";
    if (className.compareTo("Number") == 0)
      className = "java.lang.Number";
    if (className.compareTo("Short") == 0)
      className = "java.lang.Short";

    // keep the Class of the object
    aclass_ = Class.forName(className);

    // we need now to find all constructor parameters
    StringTokenizer pst = new StringTokenizer(listParams, "|");

    /**
     * fill in 2 vectors (of String)
     * vparamType: the type of the object (null if not defined)
     * vparamValue: the value of the parameters (a primitive object, an object...) 
     */
    Vector vparamType = new Vector();
    Vector vparamValue = new Vector();

    updateVector(vparamType, vparamValue, pst);

    /**
     * fill in 2 arrays (Class & Object)
     * intArgsClass: the Class type which match the type of the parameter
     * pars: the Object object which represents the value of the object
     */
    Class[] intArgsClass = new Class[vparamType.size()];
    Object[] pars = new Object[vparamValue.size()];

    updateArrays(intArgsClass, pars, vparamType, vparamValue, uObjects, sObjects);

    // build the constructor of the object and save it
    Constructor construct = aclass_.getConstructor(intArgsClass);
    obj_ = construct.newInstance(pars);
  }

  /**
   * This method returns either an Object, either a ESVectorDataSet (with VALUE flag)
   * @param line
   * @param lnb
   * @param uObjects
   * @param sObjects
   * @return The result of the method invocation
   * @throws Exception
   */
  public Object invoke(String line, int lnb, Hashtable uObjects, Hashtable sObjects)
    throws Exception {
    StringTokenizer st = new StringTokenizer(line);
    st.nextToken(); //Skip $Invoke
    logger_.debug("line=<" + line + ">");

    if (!st.hasMoreTokens())
      throw new IOException(syntaxError(line, lnb));

    String rObj = st.nextToken();

    String c =
      line
        .substring(line.indexOf(rObj) + rObj.length(), line.length())
        .trim();
    c = c.substring(1, c.length() - 1);
    int opar = c.indexOf('(');
    int dotp = c.substring(0, opar).lastIndexOf('.');
    String initObj = c.substring(0, dotp);
    String methodName = c.substring(dotp + 1, opar);
    String paramsList = c.substring(opar + 1, c.length() - 1);

    logger_.debug(
      "initObj=["
        + initObj
        + "] methodName=["
        + methodName
        + "] paramsList=["
        + paramsList
        + "]");

    StringTokenizer pst = new StringTokenizer(paramsList, "|");

    Vector vparamType = new Vector();
    Vector vparamValue = new Vector();

    updateVector(vparamType, vparamValue, pst);

    Class[] intArgsClass = new Class[vparamType.size()];
    Object[] pars = new Object[vparamValue.size()];

    updateArrays(intArgsClass, pars, vparamType, vparamValue, uObjects, sObjects);

    try {
      Method m = aclass_.getMethod(methodName, intArgsClass);
      Object r = m.invoke(obj_, pars);
      // in case we have a resulset (from a database)
      if (r instanceof ResultSet) {
          logger_.debug("r instanceof ResultSet");
          ESDBDataSet ds = new ESDBDataSet((ResultSet) r, null, null, null,null);
          return ds;
      } else if (r instanceof Collection) {
        // we need to determine if we have a Collection or not
          logger_.debug("r instanceof Collection");
        VectorDataSet vec = null;
        if (r instanceof Vector) {
          vec = new VectorDataSet(1, (Vector) r);
        } else {
          // need to transform the Collection into Vector
          Iterator iter = ((Collection) r).iterator();
          Vector newvec = new Vector();
          while (iter.hasNext()) {
            newvec.add(iter.next());
          }
          vec = new VectorDataSet(1, newvec);
        }
        return vec;
      }
      logger_.debug("r instanceof <nothing> :-)");
      if (r == null)
          logger_.debug("r is null :-(");

      return r;
      
    } catch (NoSuchMethodException nsme) {
      logger_.error("NoSuchMethodException (invoke):"+nsme + " - initObj=[" + initObj + "] methodName=[" + methodName + "] paramsList=[" + paramsList + "]");
    } catch (IllegalAccessException iae) {
      logger_.error("IllegalAccessException (invoke):"+iae + " - initObj=[" + initObj + "] methodName=[" + methodName + "] paramsList=[" + paramsList + "]");
    } catch (InvocationTargetException ite) {
      Throwable cause = ite.getCause();
      logger_.error("InvocationTargetException (invoke):"+ite + " - cause: " + cause + " initObj=[" + initObj + "] methodName=[" + methodName + "] paramsList=[" + paramsList + "]");
      if(cause != null) cause.printStackTrace(System.err);
    }
    return null;
  }

  /**
   * internal methods to fill in 2 vectors with the list of parameters
   * @param vparamType  string representation of the object type
   * @param vparamValue  string representation of the value of the object
   * @param pst  the list (StringTokenizer) of parameters
   * @param uObjects  an Hashtable which contains all the created UserObject
   */
  private void updateVector(
    Vector vparamType,
    Vector vparamValue,
    StringTokenizer pst) {

    while (pst.hasMoreTokens()) {
      // do not forget to remove blank character
      String param = pst.nextToken().trim();

      String atype = null;
      String avalue = null;

      // there is a cast !
      if (param.indexOf('(') != -1) {
        atype = param.substring(1, param.indexOf(')')).trim();
        avalue =
          param
            .substring(param.indexOf(')') + 1, param.length())
            .trim();
      } else {
        atype = null;
        avalue = param;
      }

      logger_.debug(
        "param["
          + param
          + "] atype=["
          + atype
          + "] avalue=["
          + avalue
          + "]");

      vparamType.add(atype);
      vparamValue.add(avalue);
    }
  }

  /**
   * 
   * @param intArgsClass
   * @param pars
   * @param vparamType
   * @param vparamValue
   * @param uObjects
   */
  private void updateArrays(
    Class[] intArgsClass,
    Object[] pars,
    Vector vparamType,
    Vector vparamValue,
    Hashtable uObjects,
                Hashtable sObjects) {

    try {
      for (int i = 0; i < vparamType.size(); i++) {
        String atype = (String) vparamType.get(i);
        if (atype == null) {
          /* we have 2 cases:
           * - it is a value and we need to determine the type of the object (if we can)
           * - it is an object, and we search in our UserObjects database
           */

          UserObject uo = null;
          uo = (UserObject) uObjects.get((String) vparamValue.get(i));
          if(uo == null) uo = (UserObject) sObjects.get((String) vparamValue.get(i));
          if (uo == null) {
            // it is a value
            String val = (String) vparamValue.get(i);

            if (val.indexOf('\"') != -1) {
              intArgsClass[i] = Class.forName("java.lang.String");
              vparamType.setElementAt("java.lang.String", i);
            } else {
              logger_.debug("we do nothing else...");
              // nothing, we do not know !!!!!!!
            }
          } else {
            // it is an object
            intArgsClass[i] = uo.getObjectClass();
            vparamType.setElementAt(uo.getStringType(), i);
            pars[i] = uo.getObject();
          }
        } else if (atype.compareTo("String") == 0)
          intArgsClass[i] = Class.forName("java.lang.String");
        else if (atype.compareTo("int") == 0)
          intArgsClass[i] = int.class;
        else if (atype.compareTo("boolean") == 0)
          intArgsClass[i] = boolean.class;
        else if (atype.compareTo("float") == 0)
          intArgsClass[i] = float.class;
        else if (atype.compareTo("double") == 0)
          intArgsClass[i] = double.class;
        else if (atype.compareTo("char") == 0)
          intArgsClass[i] = char.class;
        else if (atype.compareTo("byte") == 0)
          intArgsClass[i] = byte.class;
        else if (atype.compareTo("long") == 0)
          intArgsClass[i] = long.class;
        else if (atype.compareTo("short") == 0)
          intArgsClass[i] = short.class;
        else {
          logger_.debug("We have a properly defined object !!!!");
          intArgsClass[i] = Class.forName(atype);
        }
        logger_.debug("intArgsClass[i]=<" + intArgsClass[i] + ">");
      }
    } catch (ClassNotFoundException e) {
      e.printStackTrace();
    }

    for (int i = 0; i < vparamValue.size(); i++) {
      String param = (String) vparamValue.get(i);
      String atype = (String) vparamType.get(i);

      if (atype == null) {
        UserObject uo = null;
        uo = (UserObject) uObjects.get((String) vparamValue.get(i));
        if (uo != null)
          pars[i] = uo.getObject();
        else {
          uo = (UserObject) sObjects.get((String) vparamValue.get(i));
          if (uo != null)
            pars[i] = uo.getObject();
        }

      } else if ((param.indexOf("\"") == 0) && (param.lastIndexOf("\"") == param.length())) {
        pars[i] = new String(param.substring(1, param.length() - 1));
      } else if (((String) vparamType.get(i)).compareTo("int") == 0)
        pars[i] = new Integer(param);
      else if (((String) vparamType.get(i)).compareTo("boolean") == 0)
        pars[i] = new Boolean(param);
      else if (((String) vparamType.get(i)).compareTo("float") == 0)
        pars[i] = new Float(param);
      else if (((String) vparamType.get(i)).compareTo("double") == 0)
        pars[i] = new Double(param);
      else if (((String) vparamType.get(i)).compareTo("char") == 0)
        pars[i] = new Character(param.charAt(0));
      else if (((String) vparamType.get(i)).compareTo("byte") == 0)
        pars[i] = new Byte(param);
      else if (((String) vparamType.get(i)).compareTo("long") == 0)
        pars[i] = new Long(param);
      else if (((String) vparamType.get(i)).compareTo("short") == 0)
        pars[i] = new Short(param);
      else {
        logger_.debug("updateArrays: we try to get the UserObject");
        UserObject uo = null;
        if (uObjects != null)
          uo = (UserObject) uObjects.get((String) vparamValue.get(i));
        if (uo != null) {
          logger_.debug("updateArrays: uo is not null, we get it in uObjects");
          // we have to get this object to use it
          pars[i] = uo.getObject();
        }
        else {
          if (sObjects != null)
            uo = (UserObject) sObjects.get((String) vparamValue.get(i));
          if (uo != null) {
            logger_.debug("updateArrays: uo is not null, we get it in sObjects");
            // we have to get this object to use it
            pars[i] = uo.getObject();
          } else {       
            logger_.debug("updateArrays: uo is null, we only use the word as a string");
            // in this case, we do not know the object, and we use the word as a String ! 
            pars[i] =  new String(param.substring(0, param.length()));
          }
        }
      }
      logger_.debug("parse value=<" + pars[i] + ">");

    }
  }

  /* 
   * IN CASE WE WANT TO USE $ASSIGN IN ESDYNHTML.JAVA
   * WE made it in the future...
   */
  /*
    public Object invoke2(String line, int lnb) throws Exception {
  
      String c = line.substring(1, line.length()-1);
      String initObj = c.substring(0, c.indexOf('.'));
      String methodName = c.substring(c.indexOf('.')+1, c.indexOf('('));
      String paramsList = c.substring(c.indexOf('(')+1, c.length()-1);
      
      logger_.debug("initObj=["+initObj+"] methodName=["+methodName+"] paramsList=["+paramsList+"]");
      
      StringTokenizer pst = new StringTokenizer(paramsList, ",");
  
      Vector vparamType = new Vector();
      Vector vparamValue = new Vector();
      
      updateVector(vparamType, vparamValue, pst);
  
      Class[] intArgsClass = new Class[vparamType.size()];
      for (int i = 0; i<vparamType.size(); i++) {
        intArgsClass [i] = Class.forName((String)vparamType.get(i));
      }
  
      Object[] pars = new Object[vparamValue.size()];
      for (int i = 0; i<vparamValue.size(); i++) {
        String param = (String)vparamValue.get(i);
        if (param.indexOf("\"") != -1)
          pars [i] = new String(param.substring(1, param.length()-1));
      }
  
      try {
        Method m = aclass_.getMethod(methodName, intArgsClass);
        Object r = m.invoke(obj_, pars);
        //logger_.debug("************"+r.toString());
        return r;
      } catch(NoSuchMethodException nsme) {
        
      } catch(IllegalAccessException iae) {
        
      } catch(InvocationTargetException ite) {
        
      }
      return null;
    }
  
  */

  public void close() {
    logger_  = null;
    type_ = null;
    name_ = null;
    aclass_ = null;
    obj_ = null;
  }
  
  protected String getName() {
    return name_;
  }

  protected String getStringType() {
    return type_;
  }

  protected Class getObjectClass() {
    return aclass_;
  }

  protected Object getObject() {
    return obj_;
  }

  private String syntaxError(String msg, int lnb) {
    return "Syntax error line " + lnb + ": " + msg;
  }
        
  public boolean isPersist() {
    return persist_;
  }
}
